// d-lib_preset_upd.cpp
// Preset update functions
#pragma once

///////////////////////
// ENCODER RESCALING //
///////////////////////

// encoder => exp use
uint32_t enc_exp(uint64_t enc_i, uint64_t width_i, uint64_t db12_i, bool down_i) {
	uint64_t mult = db12_i << 28;
	uint64_t offs = (16 - db12_i) << 28;
	uint64_t enc_fs = enc_i << (32 - width_i);  // full scale
	uint64_t enc_mult = (enc_fs * mult) >> 32;  // * mult
	uint64_t enc_offs = enc_mult + offs;  // + offs
	double_t enc_5_27 = double_t(enc_offs) / exp2(27);  // => 5.27 decimal
	uint64_t exp_o = exp2(enc_5_27);
	if (down_i) {
		uint64_t dn = 1 << ((-2 * db12_i) + 32);
		exp_o -= dn;
	}
	return(exp_o);
}

// encoder => 16.16 inv use
uint32_t enc_inv(uint64_t enc_i, uint64_t width_i) {
	uint64_t enc_fs = enc_i << (32 - width_i);  // full scale
	uint64_t enc_frac = enc_fs >> 16;  // 16.16 fractional
	double_t enc_16_16 = double_t(enc_frac) / exp2(16);  // => 16.16 decimal
	uint64_t inv_o = exp2(16) / enc_16_16;
	return(inv_o);
}

// knee update: [0:63]48db_dn => [0:31]inv_rev
uint8_t knee_upd(uint64_t enc_i) {
	if (!enc_i) return (0);  // zero out for zero in to turn off
	if (enc_i == 63) return (31);  // max out for max in
	uint32_t use_exp = (enc_exp(enc_i, 6, 4, true)) >> 20;  // old use
	uint32_t error = 1 << 31;
	uint32_t enc_o = 0;
	for(int i=0; i<32; i++) {
		uint32_t use_inv = (enc_inv(31-i, 5) - 0xc000) >> 12;  // new use
		uint32_t error_tmp = labs(int64_t(use_exp) - int64_t(use_inv));
		if (error_tmp < error) {
			error = error_tmp;
			enc_o = i;
		}
	}
	return(enc_o);
}

// span update: [0:31]48db_dn => [0:31]inv_rev
uint8_t span_upd(uint64_t enc_i) {
	if (!enc_i) return (0);  // zero out for zero in to turn off
	if (enc_i == 31) return (31);  // max out for max in
	uint32_t use_exp = ((enc_exp(enc_i, 5, 4, true)) >> 22) + 4;  // old use (+ 2 bits res)
	uint32_t error = 1 << 31;
	uint32_t enc_o = 0;
	for(int i=0; i<32; i++) {
		uint32_t use_inv = enc_inv(31-i, 5) >> 14;  // new use (+ 2 bits res)
		uint32_t error_tmp = labs(int64_t(use_exp) - int64_t(use_inv));
		if (error_tmp < error) {
			error = error_tmp;
			enc_o = i;
		}
	}
	return(enc_o);
}

// fall update: [0:63]48db_rev => [0:63]84db_rev_dn
uint8_t fall_upd(uint64_t enc_i) {
	if (!enc_i) return (0);  // zero out for zero in to turn off
	uint32_t use_exp = (enc_exp(63-enc_i, 6, 4, false)) >> 13;  // old use
	uint32_t error = 1 << 31;
	uint32_t enc_o = 0;
	for(int i=0; i<64; i++) {
		uint32_t use_exp2 = (enc_exp(63-i, 6, 7, true)) >> 7;  // new use
		uint32_t error_tmp = labs(int64_t(use_exp) - int64_t(use_exp2));
		if (error_tmp < error) {
			error = error_tmp;
			enc_o = i;
		}
	}
	return(enc_o);
}

// rise update: [0:63]48db_rev => [0:63]72db_rev
uint8_t rise_upd(uint64_t enc_i) {
	if (!enc_i) return (0);  // zero out for zero in to turn off
	uint32_t use_exp = (enc_exp(63-enc_i, 6, 4, false)) >> 10;  // old use
	uint32_t error = 1 << 31;
	uint32_t enc_o = 0;
	for(int i=0; i<64; i++) {
		uint32_t use_exp2 = (enc_exp(63-i, 6, 6, false)) >> 7;  // new use
		uint32_t error_tmp = labs(int64_t(use_exp) - int64_t(use_exp2));
		if (error_tmp < error) {
			error = error_tmp;
			enc_o = i;
		}
	}
	return(enc_o);
}

// preset file => preset file
// read, update, write back preset file (after any significant encoder changes)
// return true if error
bool dlp_update(const string& dlp_i) {
	bool error_f = false;
	string pre_str;
	uint8_t pre_uba[PRE_PARAMS] = { 0 };
	uint8_t out_uba[PRE_PARAMS] = { 0 };
	
	// read from file
	if (!error_f) {	error_f = file_to_str(dlp_i, pre_str); }
	if (!error_f) { error_f = pre_str_to_pre_uba(pre_uba, pre_str); }

	// default: output = input
	for (uint32_t i=0; i<PRE_PARAMS; i++) {
		out_uba[i] = pre_uba[i];
	}

	////////////////////////
	// special cases here //
	////////////////////////

	// pc_p1, span, [0:31]48db_dn => [0:31]inv_rev
	out_uba[71] = span_upd(pre_uba[71]);

	// pc_p4, vloc, zero out
	out_uba[73] = 0;

	// e_p1, knee, [0:63]48db_dn => [0:31]inv_rev
	out_uba[75] = knee_upd(pre_uba[75]);

	// e_p2, fall, [0:63]48db_rev => [0:63]84db_rev_dn
	out_uba[76] = fall_upd(pre_uba[76]);

	// e_p3, rise, [0:63]48db_rev => [0:63]72db_rev
	out_uba[77] = rise_upd(pre_uba[77]);

	// e_p4, velo, range/2:
	out_uba[78] = pre_uba[78] >> 1;  

	// e_p5, damp, [0:63]48db_rev => [0:63]84db_rev_dn
	out_uba[79] = fall_upd(pre_uba[79]);

	// pp_p2, oct, constrain to [-3:3]
	if (pre_uba[83] < -3) out_uba[83] = -3;
	if (pre_uba[83] > 3) out_uba[83] = 3;

	// pp_p7, prvw, constrain to [-12:11]:
	if (pre_uba[95] < -12) out_uba[95] = -12;
	if (pre_uba[95] > 11) out_uba[95] = 11;
	
	
	// write back to file
	if (!error_f) { pre_uba_to_pre_str(out_uba, pre_str); }
	if (!error_f) { error_f = pre_str_to_dlp(pre_str, dlp_i); }
	return(error_f);
}
